//------------------------------------------------------------
// v1.0
//                    __________________
//                   |o                o|
//                  -| A0            D0 |- LED
//                  -| -             D1 |- BUS
//                  -| -             D2 |- 
//                  -| SD3           D3 |- DHT22
//                  -| SD2           D4 |- 
//                  -| SD1          3,3 |- 
//                  -| CMD          GND |- 
//                  -| SD0  nodeMCU  D5 |- 
//                  -| CLK  ESP8266  D6 |- 
//                  -| GND           D7 |- 
//                  -| 3,3           D8 |- 
//                  -| EN            RX |- 
//                  -| RST           TX |- 
//                  -| GND          GND |- 
//                  -| Vin   ____   3,3 |- 
//                   |o_____|    |_____o|
//                          |----|
//
//------------------------------------------------------------

// knihovny
#include <ESP8266WiFi.h>
#include <DHT.h>
#include <WiFiUdp.h>
#include <ArduinoOTA.h>
#include <WiFiClient.h>
#include <ESP8266WebServer.h>

// parametry
const char* ssid          = "....";              // název WiFi sítě
const char* password      = "....";         // heslo WiFi
const char* OTAhostname   = "ESP-provoz";       // název OTA
const char* host          = "milansrajer.cz";   // cílová doména
int PocetPripojeni        = 3;                  // počet pokusů o odeslání dat na server
const long interval       = 60000;              // časový interval odesílání na server (ms), 60000 = 1 minuta

// definice pinů procesoru
#define LED D0                  // vestavěná LED
#define BUS D1                  // sběrnice provozního monitoru
#define pinDHT D3               // čidlo teploty a vlhkosti DHT22

// proměnné
unsigned long preMillis = 0;    // proměnná pro časování v ms
unsigned long preMicros = 0;    // proměnná pro časování v us
String data;
int Pripojeni = PocetPripojeni; // aktuální počet realizovaných pokusů o odeslání dat na server
int TDE, TJE, TDT, T2L1, T2L2, T4L1, T4L2, SYS, ERR, T1DE, T1JE, T1DT, T2DE, T2JE, T2DT, T3DE, T3JE, T3DT, T4DE, T4JE, T4DT, JIM, AKUDE, AKUJE, AKUDT, ALA, ZAPLAVA, ZAPLAV2, KONTR;
int TDEp, TJEp, TDTp, T2L1p, T2L2p, T4L1p, T4L2p, SYSp, ERRp, T1DEp, T1JEp, T1DTp, T2DEp, T2JEp, T2DTp, T3DEp, T3JEp, T3DTp, T4DEp, T4JEp, T4DTp, JIMp, AKUDEp, AKUJEp, AKUDTp, ALAp, ZAPLAVAp, ZAPLAV2p, KONTRp;
float teplota = 0;
float vlhkost = 0;

WiFiServer server(80);          // port webserveru

#define typDHT22 DHT22          // nastavení DHT čidla
DHT mojeDHT(pinDHT, typDHT22);

//============================================================

void setup()
{

  // nastavení pinů procesoru
  pinMode(BUS, INPUT);
  pinMode(LED, OUTPUT);
  digitalWrite(LED, 1);

  // zahájení sériové komunikace
  Serial.begin(9600);
  Serial.println("");

//------------------------------------------------------------

  // připojení k WiFi
  Serial.print("pripojovani k WiFi siti ");
  Serial.println(ssid);
  WiFi.begin(ssid,password);
  
  // čekání na připojení k WiFi
  preMillis = millis();
  while (WiFi.status() != WL_CONNECTED)
  {
    if (millis() - preMillis >= 120000) {ESP.restart();}  // každé 2 minuty se ESP restartuje
    Serial.print(".");
    digitalWrite(LED, 0);     // LED blika cyklicky
    delay(50);
    digitalWrite(LED, 1);
    delay(450);
  }
  Serial.println("");

//------------------------------------------------------------

  // po připojení k WiFi
  if (WiFi.status() == WL_CONNECTED)
  { 
    Serial.print("pripojeno k WiFi, IP ");
    Serial.println(WiFi.localIP());
    
    // nastavíme ESP do režimu Station
    WiFi.mode(WIFI_AP_STA);

    // spustíme Klienta
    server.begin();
    Serial.println("Client spusten");

    // spustíme OTA
    ArduinoOTA.setHostname(OTAhostname);
    
    ArduinoOTA.onStart([]()
    {
      Serial.println("Start");
    });
    ArduinoOTA.onEnd([]()
    {
      Serial.println("\nEnd");
    });
    ArduinoOTA.onProgress([](unsigned int progress, unsigned int total)
    {
      Serial.printf("Progress: %u%%\r", (progress / (total / 100)));
    });
    ArduinoOTA.onError([](ota_error_t error)
    {
      Serial.printf("Error[%u]: ", error);
      if (error == OTA_AUTH_ERROR) Serial.println("Auth Failed");
      else if (error == OTA_BEGIN_ERROR) Serial.println("Begin Failed");
      else if (error == OTA_CONNECT_ERROR) Serial.println("Connect Failed");
      else if (error == OTA_RECEIVE_ERROR) Serial.println("Receive Failed");
      else if (error == OTA_END_ERROR) Serial.println("End Failed");
    });
    ArduinoOTA.begin();
    Serial.println("OTA spusteno");
    Serial.println("");

    // LED třikrát blikne
    digitalWrite(LED, 0);
    delay(50);
    digitalWrite(LED, 1);
    delay(50);
    digitalWrite(LED, 0);
    delay(50);
    digitalWrite(LED, 1);
    delay(50);
    digitalWrite(LED, 0);
    delay(50);
    digitalWrite(LED, 1);
    delay(250);

    // nulování počítadla
    preMillis = millis();   // právě teď začíná odpočítávání časového intervalu - cyklické odesílání dat na server
  }
}

//============================================================

void loop()
{
  ArduinoOTA.handle();  // obsluha OTA
  obsluhaKlienta();     // obsluha klienta (přímé čtení z ESP)
  
  if (digitalRead(BUS) == 0) {cteniDat();}  // pokud je sbernice v log 0, zahaj příjem dat

  if (millis() - preMillis >= interval)     // v nastavených intervalech
  {
    preMillis = millis(); // vynuluj počítadlo
    cteniSenzoru();       // načti data z DHT
    pripravData();        // připrav data pro odeslání na Endoru
    odesliNaEndoru();     // odešli na Endoru
  }
}

//============================================================

void cteniSenzoru()
{
  float teplota_nova = mojeDHT.readTemperature();   // čti teplotu
  float vlhkost_nova = mojeDHT.readHumidity();      // čti vlhkost
  
  // ošetření chyb čtení
  if ((isnan(teplota_nova)) || (isnan(vlhkost_nova)))
  {
    return;
  }

  teplota = teplota_nova;
  vlhkost = vlhkost_nova;
  return;
}

//------------------------------------------------------------

void pripravData()
{
  data = "?t=";
  data += TDE;
  data += TJE;
  data += TDT;
  data += "&t2l1=";
  data += T2L1;
  data += "&t2l2=";
  data += T2L2;
  data += "&t4l1=";
  data += T4L1;
  data += "&t4l2=";
  data += T4L2;
  data += "&sys=";
  data += SYS;
  data += "&err=";
  data += ERR;
  data += "&t1=";
  data += T1DE;
  data += T1JE;
  data += T1DT;
  data += "&t2=";
  data += T2DE;
  data += T2JE;
  data += T2DT;
  data += "&t3=";
  data += T3DE;
  data += T3JE;
  data += T3DT;
  data += "&t4=";
  data += T4DE;
  data += T4JE;
  data += T4DT;
  data += "&jim=";
  data += JIM;
  data += "&aku=";
  data += AKUDE;
  data += AKUJE;
  data += AKUDT;
  data += "&ala=";
  data += ALA;
  data += "&zap=";
  data += ZAPLAVA;
  data += "&s=";
  data += WiFi.RSSI();
  data += "&ip=";
  data += WiFi.localIP();
  data += "&teob=";
  data += teplota;
  data += "&vlob=";
  data += vlhkost;

  //vynulování chybového registru
  ERR = 128;
}

//------------------------------------------------------------

// odeslání dat na Endoru
void odesliNaEndoru()
{
  // při výpadku WiFi přeruš
  if (WiFi.status() != WL_CONNECTED) {return;}
  
  // příprava dat pro odeslání
  String url = "/----.php" + data;
  
  // připojení k serveru
  const int httpPort = 80;
  Pripojeni = PocetPripojeni;

  // návratový vektor při více pokusech o odeslání
  pripoj:

  // pokud se spojení podařilo
  WiFiClient client;
  if (client.connect(host, httpPort))
  {
    // odeslání dat
    Serial.println("odesilani");
    client.print(String("GET ") + url + " HTTP/1.1\r\n" +
      "Host: " + host + "\r\n" +
      "Accept: text/html;q=0.9,image/webp,*/*;q=0.8" +
      "Connection: close\r\n\r\n");

    // čtení ze serveru
    delay(1000);
    while (client.available())
    {
      char c = client.read();
      Serial.print(c);
    }

    // odpojení klienta
    client.stop();
    Serial.println("odeslano");

    // 2x bliknutí LED
    digitalWrite(LED, 0);
    delay(100);
    digitalWrite(LED, 1);
    delay(100);
    digitalWrite(LED, 0);
    delay(100);
    digitalWrite(LED, 1);
  }
  else
  // pokud se spojení nepodařilo
  {
    // další pokusy, případně zrušení odesílání
    if (Pripojeni != 0) {
      Serial.println("připojení se nezdařilo, další pokus");
      Pripojeni--;
      goto pripoj;
    }
    Serial.println("všechny pokusy o připojení selhaly");
    Serial.println("budou načtena nová data ze sběrnice");
  }
}

//------------------------------------------------------------
//
//  Obsluha klienta - umožňuje přímé čtení dat z ESP
//
//  192.168.___.___/xxxx
//
//  xxxx = data -> vrátí řetězec aktuálních dat
//  xxxx = sign -> vrátí sílu WiFi signálu
//  xxxx = ssid -> vrátí SSID WiFi sítě
//
//------------------------------------------------------------

void obsluhaKlienta()
{
  WiFiClient client = server.available();
  if (client)
  {
    if (client.connected())
    {
      // čtení požadavku klienta
      String req = client.readStringUntil('\r');
      Serial.print("novy klient rika: ");
      Serial.println(req);

      // separace požadavku (4 znaky po "GET /")
      String pozadavek = req.substring(5,9);
      Serial.print("pozadavek: ");
      Serial.println(pozadavek);

      String odpoved = "";

      // --- data ---
      // zpracování požadavku na kompletní data
      if (pozadavek == "data")
      {
        // příprava řetězce odpovědi
        odpoved = TDE;
        odpoved += TJE;
        odpoved += TDT;
        odpoved += "_";
        odpoved += T1DE;
        odpoved += T1JE;
        odpoved += T1DT;
        odpoved += "_";
        odpoved += T2DE;
        odpoved += T2JE;
        odpoved += T2DT;
        odpoved += "_";
        odpoved += T3DE;
        odpoved += T3JE;
        odpoved += T3DT;
        odpoved += "_";
        odpoved += T4DE;
        odpoved += T4JE;
        odpoved += T4DT;
        odpoved += "_";
        odpoved += T2L1;
        odpoved += "_";
        odpoved += T2L2;
        odpoved += "_";
        odpoved += T4L1;
        odpoved += "_";
        odpoved += T4L2;
        odpoved += "_";
        odpoved += AKUDE;
        odpoved += AKUJE;
        odpoved += AKUDT;
        odpoved += "_";
        odpoved += JIM;
        odpoved += "_";
        odpoved += SYS;
        odpoved += "_";
        odpoved += ERR;
        odpoved += "_";
        odpoved += ALA;
        odpoved += "_";
        odpoved += ZAPLAVA;
        odpoved += "_";
        odpoved += WiFi.RSSI();
        odpoved += "_";
        odpoved += WiFi.localIP();
        odpoved += "_";
        odpoved += teplota;
        odpoved += "_";
        odpoved += vlhkost;
      }

      // --- sign ---
      // zpracování požadavku na sílu signálu
      if (pozadavek == "sign")
      {
        // odpověď
        odpoved = WiFi.RSSI();
      }

      // --- ssid ---
      // zpracování požadavku na SSID WiFi sítě
      if (pozadavek == "ssid")
      {
        // odpověď
        odpoved = WiFi.SSID();
      }

      // zpracování neznámého dotazu
      if (odpoved == "")
      {
        odpoved = "neznamy pozadavek";
      }
      else
      {
        odpoved = "ESP8266_" + odpoved;
      }

      client.flush();
      delay(1);
 
      // odeslání odpovedi klientovi
      client.print(odpoved);
      delay(1000);
    }
    client.stop();
    Serial.println("klient odpojen");
    Serial.println("");
  }
}

//------------------------------------------------------------
// čtení dat ze sběrnice
//
// ESP "posluchá" komunikaci a zapisuje si data
//------------------------------------------------------------
void cteniDat()
{
  // detekce RESET impulsu (log 0 > 400us)
  preMicros = micros();
  while (micros() - preMicros <= 400)
  {
    if (digitalRead(BUS) == 1) {return;}
  }

  // po detekci reset impulsu počká na jeho ukončení (max dalších 700us)
  preMicros = micros();
  while (digitalRead(BUS) == 0)
  {
    if(micros() - preMicros >= 700) {return;}
  }

  // po detekci a ukonceni reset impulsu pokracuje ctenim adresy zarizeni...
  int ADRESA = ctiByte();

  // ... a ctenim dat z volaneho zarizeni
  digitalWrite(LED, 0); // rosvítí LED
  if (ADRESA == 0) {RidiciJednotka();} 
  if (ADRESA == 1) {KotelZdroj();}
  if (ADRESA == 2) {BojlerJimka();}
  if (ADRESA == 3) {Alarm();}
  if (ADRESA == 4) {Zaplava();}
  digitalWrite(LED, 1); // po ukončení paketu zhasne LED
}

// čtení dat z ŘÍDÍCÍ JEDNOTKY
void RidiciJednotka()
{
  TDEp = ctiByte();   //telpota v obyvaku
  TJEp = ctiByte();
  TDTp = ctiByte();
  T2L1p = ctiByte();  // T2 spodni limit (BCD)
  T2L2p = ctiByte();  // T2 horni limit (BCD)
  T4L1p = ctiByte();  // T4 spodni limit (BCD)
  T4L2p = ctiByte();  // T4 horni limit (BCD)
  SYSp = ctiByte();   // systemovy registr
  ERRp = ctiByte();   // chybovy registr
  KONTR = ctiByte();  // kontrolní byte
  if (KONTR == 85)
  {
    Serial.println("0-OK");
    TDE = TDEp;
    TJE = TJEp;
    TDT = TDTp;
    T2L1 = T2L1p;
    T2L2 = T2L2p;
    T4L1 = T4L1p;
    T4L2 = T4L2p;
    SYS = SYSp;
    ERR = ERRp;
  }
  else
  {
    Serial.println("0-X");
  }
}

// čtení dat z modulu KOTEL-ZDROJ
void KotelZdroj()
{
  T1DEp = ctiByte();
  T1JEp = ctiByte();
  T1DTp = ctiByte();
  T2DEp = ctiByte();
  T2JEp = ctiByte();
  T2DTp = ctiByte();
  AKUDEp = ctiByte();
  AKUJEp = ctiByte();
  AKUDTp = ctiByte();
  KONTR = ctiByte();
  if (KONTR == 85)
  {
    Serial.println("1-OK");
    T1DE = T1DEp;
    T1JE = T1JEp;
    T1DT = T1DTp;
    T2DE = T2DEp;
    T2JE = T2JEp;
    T2DT = T2DTp;
    AKUDE = AKUDEp;
    AKUJE = AKUJEp;
    AKUDT = AKUDTp;
  }
  else
  {
    Serial.println("1-X");
  }  
}

// čtení dat z modulu BOJLER-JÍMKA
void BojlerJimka()
{ 
  T3DEp = ctiByte();
  T3JEp = ctiByte();
  T3DTp = ctiByte();
  T4DEp = ctiByte();
  T4JEp = ctiByte();
  T4DTp = ctiByte();
  JIMp = ctiByte();
  KONTR = ctiByte();
  if (KONTR == 85)
  {
    Serial.println("2-OK");
    T3DE = T3DEp;
    T3JE = T3JEp;
    T3DT = T3DTp;
    T4DE = T4DEp;
    T4JE = T4JEp;
    T4DT = T4DTp;
    JIM = JIMp;
  }
  else
  {
    Serial.println("2-X");
  }  
}

// čtení dat z modulu ALARM
void Alarm()
{ 
  ctiByte();  // alarmy odeslane moderatorem
  ALAp = ctiByte();  // stav PGx a PGy vystupu alarmu
  KONTR = ctiByte();
  if (KONTR == 85)
  {
    Serial.println("3-OK");
    ALA = ALAp;
  }
  else
  {
    Serial.println("3-X");
  }  
}

// čtení dat z modulu ZÁPLAVA
void Zaplava()
{ 
  ZAPLAVAp = ctiByte();
  ZAPLAV2p = ctiByte();
  KONTR = ctiByte();
  if (KONTR == 85)
  {
    Serial.println("4-OK");
    Serial.println("---");
    ZAPLAVA = ZAPLAVAp;
    ZAPLAV2 = ZAPLAV2p;
  }
  else
  {
    Serial.println("4-X");
    Serial.println("---");
  }  
}

//------------------------------------------------------------

// čtení bitu
int ctiBit()
{
  // pocka na log 1 = ukonceni predchoziho bitu (max 100us, pak vrati "0")
  preMicros = micros();
  while (digitalRead(BUS) == 0)
  {
    if (micros() - preMicros >= 100) {return 0;}
  }

  // pocka na log 0 = zacatek bitu (max 100us, pak vrati "0")
  preMicros = micros();
  while (digitalRead(BUS) == 1)
  {
    if (micros() - preMicros >= 100) {return 0;}
  }

  // pocka 40us po sestupne hrane a precte hodnotu bitu
  delayMicroseconds(40);
  int x = digitalRead(BUS);

  // vrati hodnotu bitu
  return x;
}

//------------------------------------------------------------

// čtení BIN BYTE, vrací DEC
int ctiByte()
{
  int bit0 = ctiBit();
  int bit1 = ctiBit();
  int bit2 = ctiBit();
  int bit3 = ctiBit();
  int bit4 = ctiBit();
  int bit5 = ctiBit();
  int bit6 = ctiBit();
  int bit7 = ctiBit();
  int bajt = 0;
  bajt = bit0 + (bit1 * 2) + (bit2 * 4) + (bit3 * 8) + (bit4 * 16) + (bit5 * 32) + (bit6 * 64) + (bit7 * 128);
  return bajt;
}

//------------------------------------------------------------

// čtení BCD BYTE, vrací DEC
int ctiBcd()
{
  int bit0 = ctiBit();
  int bit1 = ctiBit();
  int bit2 = ctiBit();
  int bit3 = ctiBit();
  int bit4 = ctiBit();
  int bit5 = ctiBit();
  int bit6 = ctiBit();
  int bit7 = ctiBit();
  int bajt = 0;
  bajt = bit0 + (bit1 * 2) + (bit2 * 4) + (bit3 * 8) + (10 * (bit4 + (bit5 * 2) + (bit6 * 4) + (bit7 * 8)));
  return bajt;
}

//============================================================
